#import "GameSocketSocket.h"

// CocoaLumberjack
#import "DDLog.h"
#import "DDTTYLogger.h"

#include <stdlib.h>

#define GAMESOCKET_TAG_MSG 1

static int ddLogLevel = LOG_LEVEL_VERBOSE;

// Declaring "internal Properties
@interface GameSocketSocket ()
  @property (nonatomic, strong) GCDAsyncUdpSocket *socket;
  @property (nonatomic) uint16_t desiredPort;
@end

@implementation GameSocketSocket

// Public Properties
@synthesize port, serverEndpoint, serverPort, delegate;

// Internal Properties
@synthesize socket, desiredPort;

/*** Public Methods ***/

- (BOOL) start {
  if (self.desiredPort == self.port) {
    DDLogVerbose(@"Socket is already listening on %i.", self.port);
    return;
  }
  if (self.port > 0) {
    DDLogVerbose(@"Resetting socket...");
    [self reset];
  }
  DDLogVerbose(@"Starting Socket now (current port is %i)...", self.port);
  NSError *error = nil;
  NSUInteger totalTries = 3;
  NSUInteger try = 0;
  while (totalTries > try) {
    uint16_t portToTry;
    DDLogVerbose(@"The desired port is %i...", self.desiredPort);
    if (try == 0) {
      portToTry = self.desiredPort;
    } else {
      portToTry = [self.class randomPort];
    }
    DDLogVerbose(@"Trying to bind Socket to port %i...", portToTry);
    [self.socket bindToPort:portToTry error:&error];
    if (error) {
      DDLogVerbose(@"Error binding Socket to port %i: %@", portToTry, error);
    } else {
      DDLogVerbose(@"Bound Socket to port %i, starting to receive...", portToTry);
      [self.socket beginReceiving:&error];
      if (error) {
        DDLogVerbose(@"Could not begin continuous receiving: %@", error);
        return NO;
      } else {
        DDLogVerbose(@"Began continuous receiving");
        return YES;
      }
    }
    try++;
  }
  return NO;
}

- (BOOL) stop {
  DDLogVerbose(@"Stopping Socket now...");
  [self reset];
}

- (void) setPort:(uint16_t)newPort {
  if (newPort == 0) return;
  self.desiredPort = newPort;
  if (self.desiredPort == self.socket.localPort) {
    DDLogVerbose(@"Not setting port %i, because it is already set...", newPort);
    return;
  }
  if (self.isStarted) {
    DDLogVerbose(@"Switching active port from %i to %i...", self.socket.localPort, newPort);
    [self reset];
    [self start];
  }
}

- (void) sendDatagram:(GameSocketDatagram*)datagram {
  DDLogVerbose(@"Sending datagram to %@ on port %i...", datagram.endpoint, datagram.port);
  [self.socket sendData:datagram.payload toHost:datagram.endpoint port:datagram.port withTimeout:-1.0 tag:GAMESOCKET_TAG_MSG];
}

/*** Proxying to GCDAsyncUdpSocket ***/

- (uint16_t) port {
  return self.socket.localPort;
}

- (BOOL) isStarted {
  return self.port > 0;
}

/*** Internal Methods ***/

- (void) reset {
  if (self.isStarted) [self.socket close];
  self.socket = NULL;
}

+ (uint16_t) randomPort {
   return (uint16_t)(arc4random() % (49151 - 1024)) + 1024;
}

/*** Internal Getters ***/

- (GCDAsyncUdpSocket*) socket {
  if (socket) return socket;
  DDLogVerbose(@"Initializing GCDAsyncUdpSocket...");
  socket = [[GCDAsyncUdpSocket new] initWithDelegate:self delegateQueue:dispatch_get_main_queue()];
  return socket;
}

/*** Delegation Methods ***/

- (void) udpSocket:(GCDAsyncUdpSocket *)UDPSocket didReceiveData:(NSData*)data fromAddress:(NSData*)address withFilterContext:filterContext {
  DDLogVerbose(@"Received Data...");
  if ([self.delegate respondsToSelector:@selector(didReceiveDatagram:)]) {
    GameSocketDatagram *datagram = [GameSocketDatagram new];
    datagram.endpoint = [self.socket.class hostFromAddress:address];
    datagram.payload = data;
    [self.delegate didReceiveDatagram:datagram];
  }
}

@end
